from datetime import datetime, timedelta

import dateutil
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal
from textual.reactive import Reactive, reactive
from textual.widgets import Button, Label, Static

from bagels.components.indicators import EmptyIndicator
from bagels.components.modules.spending.plots import (
    BalancePlot,
    SpendingPlot,
    SpendingTrajectoryPlot,
)
from bagels.components.tplot import PlotextPlot
from bagels.components.tplot.plot import rgbify_hex
from bagels.config import CONFIG
from bagels.managers.utils import get_start_end_of_period
from bagels.utils.format import format_period_to_readable


class Spending(Static):
    PLOT_TYPES = [SpendingTrajectoryPlot, SpendingPlot, BalancePlot]

    can_focus = True
    periods: Reactive[int] = reactive(1)
    current_plot: Reactive[int] = reactive(0)

    BINDINGS = [
        Binding("+", "zoom_in", "Zoom in", show=True),
        Binding("-", "zoom_out", "Zoom out", show=True),
    ]

    def __init__(self, page_parent: Static, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="spending-container", classes="module-container"
        )
        super().__setattr__("border_title", "Spending")
        self._plots = [plot_cls(self.app) for plot_cls in self.PLOT_TYPES]
        self.page_parent = page_parent

    def on_mount(self) -> None:
        self.rebuild()
        self.app.theme_changed_signal2.subscribe(self, lambda theme: self.rebuild())
        self.query(Button)[4].focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "inc-offset":
            self.page_parent.action_inc_offset()
        elif event.button.id == "dec-offset":
            self.page_parent.action_dec_offset()
        elif event.button.id == "zoom-in":
            self.action_zoom_in()
        elif event.button.id == "zoom-out":
            self.action_zoom_out()
        elif event.button.id.startswith("plot-"):
            try:
                plot_index = int(event.button.id.split("-")[1])
                self.current_plot = plot_index
                self.query_one(".selected").set_classes("")
                event.button.set_classes("selected")
                self.rebuild()
            except ValueError:
                pass

    def rebuild(self) -> None:
        empty = self.query_one(EmptyIndicator)
        plotext = self.query_one(PlotextPlot)
        zoom_in_button = self.query_one("#zoom-in")
        zoom_out_button = self.query_one("#zoom-out")
        label = self.query_one(".current-view-label")
        plotext.display = False  # make plotext update by toggling display... for some reason. Maybe a bug? Who knows.
        plot = self._plots[self.current_plot]

        start_of_period, end_of_period = get_start_end_of_period(
            self.page_parent.offset, "month"
        )
        self.app.log(
            f"The plot type {plot.name} has support for cross periods: {plot.supports_cross_periods}"
        )

        # ----------- Cross period ----------- #

        self.app.log(
            f'Getting spending trend from "{start_of_period}" to "{end_of_period} ({self.page_parent.offset} months ago)"'
        )
        string = format_period_to_readable(
            {"offset": self.page_parent.offset, "offset_type": "month"}
        )

        if plot.supports_cross_periods:
            start_of_period = start_of_period - dateutil.relativedelta.relativedelta(
                months=self.periods - 1
            )
            zoom_in_button.display = True
            zoom_out_button.display = True
            if self.periods == 1:
                label.update(string)
            else:
                label.update(
                    f"{string} to {format_period_to_readable({'offset': self.page_parent.offset - self.periods + 1, 'offset_type': 'month'})}"
                )
        else:
            zoom_in_button.display = False
            zoom_out_button.display = False
            label.update(string)

        plt = self.query_one(PlotextPlot).plt
        plt.clear_data()
        plt.clear_figure()

        # ------------- get data ------------- #

        data = plot.get_data(start_of_period, end_of_period)
        total_days = (
            end_of_period - start_of_period
        ).days + 1  # add one to include the end date
        correct_data = len(data) > 0
        empty.display = not correct_data
        plotext.display = correct_data
        if not correct_data:
            return

        # --------------- plot --------------- #

        bagel_theme = self.app.themes[self.app.app_theme]

        def get_theme_color(key):
            return (
                rgbify_hex(getattr(bagel_theme, key, None))
                if getattr(bagel_theme, key, None) is not None
                else rgbify_hex(self.app.theme_variables[key])
            )

        dates = [
            (start_of_period + timedelta(days=i)).strftime("%d/%m/%Y")
            for i in range(total_days)
        ]

        plot.plot(
            plt,
            start_of_period,
            end_of_period,
            self.page_parent.offset,
            data,
            dates,
            get_theme_color,
        )

        plt.plot(
            dates,
            data,
            marker=CONFIG.defaults.plot_marker,
            color=get_theme_color("accent"),
        )

        # -------------- styling ------------- #

        plt.date_form(output_form="d")
        plt.xfrequency(total_days)
        plt.xlim(dates[0], dates[-1])

        today = datetime.now()
        line_color = get_theme_color("panel")
        fdow_line_color = get_theme_color("secondary")
        today_line_color = get_theme_color("success")

        if plot.supports_cross_periods and self.periods > 1:

            def compare_function(dd):
                return dd.day == 1
        else:

            def compare_function(dd):
                return dd.weekday() == CONFIG.defaults.first_day_of_week

        for d in dates:
            d_datetime = datetime.strptime(d, "%d/%m/%Y")
            if d_datetime.date() == today.date():
                plt.vline(d, today_line_color)
            elif compare_function(d_datetime):
                plt.vline(d, fdow_line_color)
            else:
                plt.vline(d, line_color)

        plt.xaxes(False)
        plt.yaxes(False)

    def check_supports_cross_periods(self) -> bool:
        if not self._plots[self.current_plot].supports_cross_periods:
            self.app.notify(
                "Current plot does not this operation.",
                title="Cross periods not supported",
                severity="information",
            )
            return False
        return True

    def action_zoom_in(self) -> None:
        if not self.check_supports_cross_periods():
            return
        self.periods = self.periods - 1 if self.periods > 1 else 1
        self.rebuild()

    def action_zoom_out(self) -> None:
        if not self.check_supports_cross_periods():
            return
        self.periods = self.periods + 1 if self.periods < 12 else 12
        self.rebuild()

    def compose(self) -> ComposeResult:
        with Horizontal(id="top-controls-container"):
            yield Button("+", id="zoom-in")
            yield Button("<<<", id="dec-offset")
            yield Label("UPDATEME", classes="current-view-label")
            yield Button(">>>", id="inc-offset")
            yield Button("-", id="zoom-out")
        yield PlotextPlot()
        yield EmptyIndicator("No data to display")
        with Horizontal(id="bottom-controls-container"):
            for i, plot in enumerate(self._plots):
                button_classes = "selected" if i == self.current_plot else ""
                yield Button(plot.name, id=f"plot-{i}", classes=button_classes)
                if i != len(self._plots) - 1:
                    yield Label(" | ")



from datetime import datetime, timedelta

from sqlalchemy import func
from sqlalchemy.orm import joinedload, sessionmaker

from bagels.managers.splits import create_split, get_splits_by_record_id, update_split
from bagels.managers.utils import get_operator_amount, get_start_end_of_period
from bagels.models.account import Account
from bagels.models.category import Category
from bagels.models.database.app import db_engine
from bagels.models.record import Record
from bagels.models.split import Split

Session = sessionmaker(bind=db_engine)


# region Create
def create_record(record_data: dict):
    session = Session()
    try:
        record = Record(**record_data)
        session.add(record)
        session.commit()
        session.refresh(record)
        session.expunge(record)
        return record
    finally:
        session.close()


def create_record_and_splits(record_data: dict, splits_data: list[dict]):
    session = Session()
    try:
        record = create_record(record_data)
        for split in splits_data:
            split["recordId"] = record.id
            create_split(split)
        return record
    finally:
        session.close()


# region Get
def get_record_by_id(record_id: int, populate_splits: bool = False):
    session = Session()
    try:
        query = session.query(Record).options(
            joinedload(Record.category), joinedload(Record.account)
        )

        if populate_splits:
            query = query.options(
                joinedload(Record.splits).options(
                    joinedload(Split.account), joinedload(Split.person)
                )
            )

        record = query.get(record_id)
        return record
    finally:
        session.close()


def get_record_total_split_amount(record_id: int):
    session = Session()
    try:
        splits = get_splits_by_record_id(record_id)
        return sum(split.amount for split in splits)
    finally:
        session.close()


def get_records(
    offset: int = 0,
    offset_type: str = "month",
    account_id: int = None,
    category_piped_names: str = None,
    operator_amount: str = None,
    label: str = None,
):
    session = Session()
    try:
        query = session.query(Record).options(
            joinedload(Record.category),
            joinedload(Record.account),
            joinedload(Record.transferToAccount),
            joinedload(Record.splits).options(
                joinedload(Split.account), joinedload(Split.person)
            ),
        )

        start_of_period, end_of_period = get_start_end_of_period(offset, offset_type)
        query = query.filter(
            Record.date >= start_of_period, Record.date < end_of_period
        )

        if account_id not in [None, ""]:
            query = query.filter(Record.accountId == account_id)
        if category_piped_names not in [None, ""]:
            category_names = category_piped_names.split("|")
            query = query.join(Record.category).filter(
                Category.name.in_(category_names)
            )
        if operator_amount not in [None, ""]:
            operator, amount = get_operator_amount(operator_amount)
            if operator and amount:
                query = query.filter(Record.amount.op(operator)(amount))
        if label not in [None, ""]:
            query = query.filter(Record.label.ilike(f"%{label}%"))

        createdAt_column = getattr(Record, "createdAt")
        date_column = func.date(getattr(Record, "date"))
        query = query.order_by(date_column.desc(), createdAt_column.desc())

        records = query.all()
        return records
    finally:
        session.close()


def _get_spending_records(session, start_date, end_date):
    """Common function to fetch records for spending calculations"""
    return (
        session.query(Record)
        .filter(
            Record.isIncome == False,  # noqa: E712
            Record.date >= start_date,
            Record.date < end_date,
            Record.isTransfer == False,  # noqa: E712
        )
        .options(joinedload(Record.splits))
        .all()
    )


def _calculate_daily_spending(records, start_date, end_date, cumulative=False):
    """Calculate daily spending with optional cumulative sum"""
    daily_spending = {}
    for record in records:
        date_key = record.date.date()
        splits_sum = sum(split.amount for split in record.splits)
        actual_spend = record.amount - splits_sum
        daily_spending[date_key] = daily_spending.get(date_key, 0) + actual_spend

    current_date = start_date.date()
    end_date_normalized = end_date.date()
    today = datetime.today().date()
    result = []
    running_total = 0

    while current_date <= end_date_normalized:
        if current_date <= today:
            daily_amount = daily_spending.get(current_date, 0)
            if cumulative:
                running_total += daily_amount
                result.append(running_total)
            else:
                result.append(daily_amount)
        current_date += timedelta(days=1)

    return result


def get_spending(start_date, end_date) -> list[float]:
    """Gets a list of spent amounts for each day in the period, less split amounts of the records"""
    session = Session()
    try:
        records = _get_spending_records(session, start_date, end_date)
        return _calculate_daily_spending(
            records, start_date, end_date, cumulative=False
        )
    finally:
        session.close()


def get_spending_trend(start_date, end_date) -> list[float]:
    """Gets a cumulative list of spent amounts for each day in the period"""
    session = Session()
    try:
        records = _get_spending_records(session, start_date, end_date)
        return _calculate_daily_spending(records, start_date, end_date, cumulative=True)
    finally:
        session.close()


def is_record_all_splits_paid(record_id: int):
    session = Session()
    try:
        splits = get_splits_by_record_id(record_id)
        return all(split.isPaid for split in splits)
    finally:
        session.close()


def get_daily_balance(start_date, end_date) -> list[float]:
    """Gets a list of account balances for each day in the period"""
    # Calculating net beginning balance

    # Handle records up till start_date, then process each day one by one.

    # Records.isIncome ? add : subtract

    # for Record.splits => Record.isIncome ? subtract Split.amount : add Split.amount

    # Record.isTransfer ? Record.transferToAccount.name == "Outside source" ? subtract : Record.account.name == "Outside source" ? add : ignore

    # return result
    session = Session()
    try:
        accounts = session.query(Account).filter(Account.deletedAt.is_(None)).all()
        total_balance = sum(a.beginningBalance for a in accounts)
        old_records = (
            session.query(Record)
            .filter(
                Record.date < start_date, Record.accountId.in_([a.id for a in accounts])
            )
            .options(
                joinedload(Record.splits),
                joinedload(Record.account),
                joinedload(Record.transferToAccount),
            )
            .all()
        )

        def adjust_balance(r):
            if r.isTransfer:
                if r.transferToAccount and r.transferToAccount.name == "Outside source":
                    return -r.amount
                if r.account and r.account.name == "Outside source":
                    return r.amount
                return 0
            if r.isIncome:
                return r.amount - sum(s.amount for s in r.splits)
            return -r.amount + sum(s.amount for s in r.splits)

        for rec in old_records:
            total_balance += adjust_balance(rec)

        results = []
        current = start_date
        today = datetime.today()
        while current <= end_date:
            if current > today:
                break
            day_records = (
                session.query(Record)
                .filter(
                    func.date(Record.date) == current.date(),
                    Record.accountId.in_([a.id for a in accounts]),
                )
                .options(
                    joinedload(Record.splits),
                    joinedload(Record.account),
                    joinedload(Record.transferToAccount),
                )
                .all()
            )
            day_effect = sum(adjust_balance(dr) for dr in day_records)
            total_balance += day_effect
            results.append(total_balance)
            current += timedelta(days=1)
        return results
    finally:
        session.close()


# region Update
def update_record(record_id: int, updated_data: dict):
    session = Session()
    try:
        record = session.query(Record).get(record_id)
        if record:
            for key, value in updated_data.items():
                setattr(record, key, value)
            session.commit()
            session.refresh(record)
            session.expunge(record)
        return record
    finally:
        session.close()


def update_record_and_splits(
    record_id: int, record_data: dict, splits_data: list[dict]
):
    session = Session()
    try:
        record = update_record(record_id, record_data)
        record_splits = get_splits_by_record_id(record_id)
        for index, split in enumerate(record_splits):
            update_split(split.id, splits_data[index])
        return record
    finally:
        session.close()


# region Delete
def delete_record(record_id: int):
    session = Session()
    try:
        record = session.query(Record).get(record_id)
        if record:
            session.delete(record)
            session.commit()
        return record
    finally:
        session.close()



from pathlib import Path
from time import sleep

import click
from rich.progress import (
    BarColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)

from bagels.locations import config_file, database_file, set_custom_root
from bagels.versioning import get_current_version, get_pypi_version, needs_update


@click.group(invoke_without_command=True)
@click.option(
    "--at",
    type=click.Path(exists=True, file_okay=True, dir_okay=True, path_type=Path),
    help="Specify the path.",
)
@click.option(
    "--migrate",
    type=click.Choice(["actualbudget"]),
    help="Specify the migration type.",
)
@click.option(
    "--source",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=Path),
    help="Path to source database file for migration.",
)
@click.pass_context
def cli(ctx, at: Path | None, migrate: str | None, source: Path | None):
    """Bagels CLI."""
    if at:
        set_custom_root(at)

    if migrate:
        if not source:
            raise click.UsageError("--source is required when using --migrate")

        if migrate == "actualbudget":
            try:
                click.echo(f"Starting migration from {source}")
                from bagels.models.database.app import init_db

                init_db()

                from bagels.locations import database_file
                from bagels.migrations.migrate_actualbudget import (
                    BudgetToBagelsMigration,
                )

                migrator = BudgetToBagelsMigration(str(source), str(database_file()))
                migrator.migrate()
                click.echo(click.style("Migration completed successfully!", fg="green"))
                return
            except Exception as e:
                click.echo(click.style(f"Migration failed: {str(e)}", fg="red"))
                ctx.exit(1)

    if ctx.invoked_subcommand is None:
        with Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            transient=True,
        ) as progress:
            task = progress.add_task(f"Loading configuration from '{at}'...", total=3)

            from bagels.config import load_config

            load_config()

            from bagels.config import CONFIG

            if CONFIG.state.check_for_updates:
                progress.update(task, advance=1, description="Checking for updates...")

                if needs_update():
                    new = get_pypi_version()
                    cur = get_current_version()
                    click.echo(
                        click.style(
                            f"New version available ({cur} -> {new})! Update with:",
                            fg="yellow",
                        )
                    )
                    click.echo(click.style("```uv tool upgrade bagels```", fg="cyan"))
                    click.echo(
                        click.style(
                            "You can disable this check in-app using the command palette.",
                            fg="bright_black",
                        )
                    )
                    sleep(2)

            progress.update(task, advance=1, description="Initializing database...")

            from bagels.models.database.app import init_db

            init_db()
            progress.update(task, advance=1, description="Starting application...")

            from bagels.app import App

            app = App()
            progress.update(task, advance=1)

        app.run()


@cli.command()
@click.argument("thing_to_locate", type=click.Choice(["config", "database"]))
def locate(thing_to_locate: str) -> None:
    if thing_to_locate == "config":
        print("Config file:")
        print(config_file())
    elif thing_to_locate == "database":
        print("Database file:")
        print(database_file())


if __name__ == "__main__":
    cli()



from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.widgets import Button, Input, Label, Static

from bagels.config import CONFIG, write_state
from bagels.managers.utils import (
    dynamic_cache,
    get_income_to_use,
    get_period_figures,
    try_method_query_one,
)
from bagels.models.category import Nature


class Budgets(Static):
    can_focus = True

    def __init__(self, page_parent, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="budgets-container", classes="module-container"
        )
        super().__setattr__("border_title", "Budgeting")
        self.page_parent = page_parent

    # --------------- Hooks -------------- #

    def on_mount(self) -> None:
        self.rebuild()

    def _write_state(self, key: str, value: str, typing=None) -> bool:
        current_value = CONFIG.state
        for part in key.split("."):
            current_value = getattr(current_value, part)
        if value != "":
            typed_val = value
            if typing:
                typed_val = typing(value)
            if current_value != typed_val:
                write_state(key, typed_val)
            return True
        return False

    def on_input_changed(self, event: Input.Changed) -> None:
        updated = False
        if event.input.id == "savings-input":
            if self.savings_assess_metric.startswith("percentage"):
                updated = self._write_state(
                    "budgeting.savings_percentage", event.value, float
                )
            else:
                updated = self._write_state(
                    "budgeting.savings_amount", event.value, float
                )

        elif event.input.id == "wants-input":
            if self.wants_spending_assess_metric.startswith("percentage"):
                updated = self._write_state(
                    "budgeting.wants_spending_percentage", event.value, float
                )
            else:
                updated = self._write_state(
                    "budgeting.wants_spending_amount", event.value, float
                )
        if updated:
            self.rebuild()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id.startswith("savings-"):
            metric = event.button.id.replace("savings-", "")
            self._write_state("budgeting.savings_assess_metric", metric)
        elif event.button.id.startswith("wants-"):
            metric = event.button.id.replace("wants-", "")
            self._write_state("budgeting.wants_spending_assess_metric", metric)
        self.rebuild()

    # region Builders
    # ------------- Builders ------------- #

    def rebuild(self) -> None:
        self.savings_assess_metric = CONFIG.state.budgeting.savings_assess_metric
        try_method_query_one(self, "#savings-row > .selected", "set_classes", [""])
        try_method_query_one(
            self, f"#savings-{self.savings_assess_metric}", "set_classes", ["selected"]
        )
        restrict_percentage = r"0\.([0-9]){0,2}"
        restrict_amount = r"\d*\.?\d{0,2}"
        input = self.query_one("#savings-input")
        if self.savings_assess_metric.startswith("percentage"):
            input.value = str(CONFIG.state.budgeting.savings_percentage)
            input.restrict = restrict_percentage
        else:
            input.value = str(CONFIG.state.budgeting.savings_amount)
            input.restrict = restrict_amount

        self.wants_spending_assess_metric = (
            CONFIG.state.budgeting.wants_spending_assess_metric
        )
        input = self.query_one("#wants-input")
        try_method_query_one(self, "#wants-row > .selected", "set_classes", [""])
        try_method_query_one(
            self,
            f"#wants-{self.wants_spending_assess_metric}",
            "set_classes",
            ["selected"],
        )
        if self.wants_spending_assess_metric.startswith("percentage"):
            input.value = str(CONFIG.state.budgeting.wants_spending_percentage)
            input.restrict = restrict_percentage
        else:
            input.value = str(CONFIG.state.budgeting.wants_spending_amount)
            input.restrict = restrict_amount
        self._rebuild_income_bar()

    def _rebuild_income_bar(self) -> None:
        offset = self.page_parent.offset
        net_income = dynamic_cache(get_income_to_use, offset)
        net_expenses = dynamic_cache(
            get_period_figures, isIncome=False, offset=offset, offset_type="month"
        )
        amount_to_save = round(
            net_income * CONFIG.state.budgeting.savings_percentage
            if self.savings_assess_metric.startswith("percentage")
            else CONFIG.state.budgeting.savings_amount,
            CONFIG.defaults.round_decimals,
        )
        want_quota = round(
            (net_income - amount_to_save)
            * CONFIG.state.budgeting.wants_spending_percentage
            if self.wants_spending_assess_metric.startswith("percentage")
            else CONFIG.state.budgeting.wants_spending_amount,
            CONFIG.defaults.round_decimals,
        )
        expenses_must = round(
            dynamic_cache(
                get_period_figures,
                isIncome=False,
                offset=offset,
                offset_type="month",
                nature=Nature.MUST,
            ),
            CONFIG.defaults.round_decimals,
        )
        expenses_need = round(
            dynamic_cache(
                get_period_figures,
                isIncome=False,
                offset=offset,
                offset_type="month",
                nature=Nature.NEED,
            ),
            CONFIG.defaults.round_decimals,
        )
        expenses_want = round(
            net_expenses - expenses_must - expenses_need,
            CONFIG.defaults.round_decimals,
        )

        self.app.log(
            f"Data: net_income={net_income}, net_expenses={net_expenses}, amount_to_save={amount_to_save}, expenses_must={expenses_must}, expenses_need={expenses_need}, expenses_want={expenses_want}"
        )

        income_bar_container = self.query_one("#income-bar")
        empty_bar = self.query_one(".empty-bar")

        income_bar_container.display = not not net_income
        empty_bar.display = not net_income

        if not net_income:
            return

        row1 = income_bar_container.query_one("#row-1")
        p_expenses = round(net_expenses / net_income * 100)
        p_saving = round(amount_to_save / net_income * 100)
        row1_columns = f"{p_expenses}% {p_saving}% 1fr"
        self.app.log(row1_columns)
        row1.styles.grid_columns = row1_columns

        row2 = income_bar_container.query_one("#row-2")
        row2.styles.grid_columns = f"{p_expenses}% 1fr"

        label_spent_amount = income_bar_container.query_one("#label-spent-amount")
        label_spent_amount.update(str(net_expenses))
        label_save_amount = income_bar_container.query_one("#label-save-amount")
        label_save_amount.update(str(amount_to_save))
        label_remaining_amount = income_bar_container.query_one(
            "#label-remaining-amount"
        )
        remaining = round(
            net_income - net_expenses - amount_to_save, CONFIG.defaults.round_decimals
        )
        label_remaining_amount.update(str(remaining))

        row3 = income_bar_container.query_one("#row-3")
        row3.display = not not net_expenses  # only display if there are expenses

        row4 = income_bar_container.query_one("#row-4")
        row4.display = not not net_expenses

        row5 = income_bar_container.query_one("#row-5")
        row5.display = not not net_expenses

        if not not net_expenses:
            p_tospend = round(100 - p_saving)
            row3.styles.width = f"{p_tospend}%"

            p_quota = round(100 - p_saving - p_expenses)
            p_must = round(expenses_must / net_expenses * p_expenses)
            p_need = round(expenses_need / net_expenses * p_expenses)
            p_want = round(expenses_want / net_expenses * p_expenses)
            row4_columns = f"{p_must}% {p_need}% {p_quota}% {p_want}% 1fr"
            self.app.log(row4_columns)
            row4.styles.grid_columns = row4_columns

            row3.styles.grid_columns = f"{p_must}% {p_need}% 1fr"
            label_spent_must_amount = row3.query_one("#label-spent-must-amount")
            label_spent_must_amount.update(str(expenses_must))
            label_spent_need_amount = row3.query_one("#label-spent-need-amount")
            label_spent_need_amount.update(str(expenses_need))
            label_spent_want_amount = row3.query_one("#label-want-remaining-amount")
            label_spent_want_amount.update(f"{str(expenses_want)} / {str(want_quota)}")

            row5.styles.width = f"{p_tospend}%"
            bar_want_quota = row5.query_one("#bar-must-quota")
            p_want_quota = round(want_quota / (net_income - amount_to_save) * 100)
            bar_want_quota.styles.width = f"{p_want_quota}%"

    # region View
    # --------------- View --------------- #
    def compose(self) -> ComposeResult:
        with Horizontal(id="savings-row", classes="config-rows"):
            yield Label("Savings: ", id="savings-label")
            yield Button(
                "% Period Income",
                tooltip="Save a percentage of this period's income. Will temporarily use last period's value if income not yet recorded this period. Will use fallback value otherwise.\n\nValue: \"percentagePeriodIncome\"",
                id="savings-percentagePeriodIncome",
            )
            yield Button(
                "Set amount",
                tooltip='Save a set amount each period.\n\nValue: "setAmount"',
                id="savings-setAmount",
            )
            yield Input(placeholder="0.0", id="savings-input")
        with Horizontal(id="wants-row", classes="config-rows"):
            yield Label("Wants: ", id="wants-label")
            yield Button(
                "% Quota",
                tooltip='Limit spending on WANTs each period by a percentage of the amount left after savings.\n\nValue: "percentage"',
                id="wants-percentageQuota",
            )
            yield Button(
                "Set amount",
                tooltip='Limit spending on WANTs each period by a set amount.\n\nValue: "setAmount"',
                id="wants-setAmount",
            )
            yield Input(
                placeholder="0.0",
                id="wants-input",
            )
        with Container(id="income-bar"):
            with Container(id="row-0"):
                with Horizontal():
                    yield Label("╭ Spent: ")
                    yield Label("0.0", id="label-spent-amount")
                with Horizontal(id="remaining"):
                    yield Label("Remaining: ")
                    yield Label("0.0", id="label-remaining-amount")
                    yield Label(" ╮")
            with Container(id="row-1"):
                yield Static(id="bar-spent")
                yield Static(id="bar-saving")
                yield Static(id="bar-remaining")
            with Container(id="row-2"):
                yield Static()
                with Horizontal():
                    yield Label("╰ Save: ")
                    yield Label("0.0", id="label-save-amount")
            with Container(id="row-3"):
                with Horizontal():
                    yield Label("╭ Must: ")
                    yield Label("0.0", id="label-spent-must-amount")
                with Horizontal():
                    yield Label("╭ Need: ")
                    yield Label("0.0", id="label-spent-need-amount")
                with Horizontal(id="want"):
                    yield Label("Want: ")
                    yield Label("(0.0 / 0.0)", id="label-want-remaining-amount")
                    yield Label(" ╮")
            with Container(id="row-4"):
                yield Static(id="bar-spent-must")
                yield Static(id="bar-spent-need")
                yield Static(id="bar-spent-quota")
                yield Static(id="bar-spent-want")
                yield Label("Saved")
            with Container(id="row-5"):
                yield Static(id="bar-must-quota")
        with Container(classes="empty-bar"):
            yield Label("Nothing to display. Budgeting depends on your income!")



from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Container
from textual.widgets import Static

from bagels.components.bagel import Bagel
from bagels.components.modules.budgets import Budgets
from bagels.components.modules.categories import Categories
from bagels.components.modules.people import People
from bagels.components.modules.spending import Spending
from bagels.managers.accounts import get_accounts_count
from bagels.managers.categories import get_categories_count


class Manager(Static):
    offset = 0

    BINDINGS = [
        Binding("left", "dec_offset", "Shift back", show=True),
        Binding("right", "inc_offset", "Shift front", show=True),
    ]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, id="manager-page")
        self.isReady = get_accounts_count() and get_categories_count()
        self.spendings_module = Spending(page_parent=self)
        self.categories_module = Categories()
        self.budgets_module = Budgets(page_parent=self)
        self.people_module = People()

    def on_mount(self) -> None:
        pass

    # -------------- Helpers ------------- #

    def rebuild(self) -> None:
        if self.isReady:
            self.spendings_module.rebuild()
            self.categories_module.rebuild()
            self.budgets_module.rebuild()
            self.people_module.rebuild()

    # region Callbacks
    # ------------- Callbacks ------------ #

    def action_inc_offset(self) -> None:
        if self.offset < 0:
            self.offset += 1
            self.spendings_module.rebuild()
            self.budgets_module.rebuild()

    def action_dec_offset(self) -> None:
        self.offset -= 1
        self.spendings_module.rebuild()
        self.budgets_module.rebuild()

    # region View
    # --------------- View --------------- #

    def compose(self) -> ComposeResult:
        if self.isReady:
            with Static(classes="manager-modules-container"):
                with Container(classes="left"):
                    yield self.spendings_module
                    yield self.budgets_module
                with Container(classes="right"):
                    yield self.categories_module
                    yield self.people_module
        else:
            with Center():
                yield Bagel()



from typing import ClassVar

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container
from textual.events import DescendantBlur, DescendantFocus
from textual.reactive import reactive
from textual.widgets import Button, Input, Static, Switch

from bagels.components.datatable import DataTable
from bagels.components.indicators import EmptyIndicator
from bagels.components.modules.records._cud import RecordCUD
from bagels.components.modules.records._table_builder import RecordTableBuilder
from bagels.config import CONFIG
from bagels.forms.person_forms import PersonForm


class DisplayMode:
    DATE = "d"
    PERSON = "p"


class Records(RecordCUD, RecordTableBuilder, Static):
    DEFAULT_CSS = """\
Records .label-highlight-match {
    color: $accent-lighten-2;
    text-style: bold;
}
    """

    COMPONENT_CLASSES: ClassVar[set[str]] = {
        "label-highlight-match",
    }

    BINDINGS = [
        (CONFIG.hotkeys.new, "new", "Add"),
        (CONFIG.hotkeys.delete, "delete", "Delete"),
        (CONFIG.hotkeys.edit, "edit", "Edit"),
        (CONFIG.hotkeys.home.new_transfer, "new_transfer", "Transfer"),
        (CONFIG.hotkeys.home.toggle_splits, "toggle_splits", "Toggle Splits"),
        Binding(
            CONFIG.hotkeys.home.display_by_person,
            "display_by_person",
            "Display by Person",
            show=False,
        ),
        Binding(
            CONFIG.hotkeys.home.display_by_date,
            "display_by_date",
            "Display by Date",
            show=False,
        ),
    ]

    can_focus = True
    show_splits = True
    displayMode = reactive(DisplayMode.DATE)
    FILTERS = {}
    FILTER_LABEL_TIPS = {
        "filter-category": "Shopping|Food|Dining",
        "filter-amount": ">=123.45",
        "filter-label": "Dinner with friends",
    }

    def __init__(self, parent: Static, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="records-container", classes="module-container"
        )
        super().__setattr__("border_title", "Records")
        self.page_parent = parent
        self.person_form = PersonForm()
        self.FILTERS = {
            "category": lambda: self.query_one("#filter-category").value,
            "amount": lambda: self.query_one("#filter-amount").value,
            "label": lambda: self.query_one("#filter-label").value,
            "enabled": lambda: self.query_one("#toggle-filter").value,
        }

    def on_mount(self) -> None:
        self.rebuild()

    # region Callbacks
    # ------------- Callbacks ------------ #

    def on_data_table_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        current_row_index = event.cursor_row
        if event.row_key and event.row_key.value:
            self.current_row = event.row_key.value
            self.current_row_index = current_row_index
        else:
            self.current_row = None
            self.current_row_index = None

    def watch_displayMode(self, displayMode: DisplayMode) -> None:
        self.query_one("#display-date").classes = (
            "selected" if displayMode == DisplayMode.DATE else ""
        )
        self.query_one("#display-person").classes = (
            "selected" if displayMode == DisplayMode.PERSON else ""
        )

    def action_toggle_splits(self) -> None:
        self.show_splits = not self.show_splits
        self.rebuild()

    def action_display_by_person(self) -> None:
        self.displayMode = DisplayMode.PERSON
        self.rebuild()

    def action_display_by_date(self) -> None:
        self.displayMode = DisplayMode.DATE
        self.rebuild()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        match event.button.id:
            case "prev-month":
                self.page_parent.action_prev_month()
            case "next-month":
                self.page_parent.action_next_month()
            case "display-date":
                self.action_display_by_date()
            case "display-person":
                self.action_display_by_person()
            case _:
                pass

    def on_input_changed(self, event: Input.Changed) -> None:
        if self.FILTERS["enabled"]():
            self.rebuild(focus=False)

    def on_switch_changed(self, event: Switch.Changed) -> None:
        self.rebuild(focus=False)

    def on_descendant_focus(self, event: DescendantFocus) -> None:
        if event.widget.id.startswith("filter-"):
            event.widget.placeholder = self.FILTER_LABEL_TIPS[event.widget.id]

    def on_descendant_blur(self, event: DescendantBlur) -> None:
        if event.widget.id.startswith("filter-"):
            event.widget.placeholder = "Filter " + event.widget.id.split("-")[1]

    # region View
    # --------------- View --------------- #

    def compose(self) -> ComposeResult:
        with Container(classes="selectors"):
            with Container(id="display-selector"):
                yield Button(
                    f"Date ({CONFIG.hotkeys.home.display_by_date})", id="display-date"
                )
                yield Button(
                    f"Person ({CONFIG.hotkeys.home.display_by_person})",
                    id="display-person",
                )
            with Container(classes="filtering", id="filter-container"):
                yield Input(id="filter-category", placeholder="Filter category")
                yield Input(
                    id="filter-amount",
                    placeholder="Filter amount",
                    restrict=r"^(>=|>|=|<=|<)?\d*\.?\d*$",
                )
                yield Input(id="filter-label", placeholder="Filter label")
                yield Switch(id="toggle-filter", animate=False)
        self.table = DataTable(
            id="records-table",
            cursor_type="row",
            cursor_foreground_priority=True,
            zebra_stripes=True,
        )
        yield self.table
        yield EmptyIndicator("No entries")



from datetime import timedelta

from rich.text import Text

from bagels.components.datatable import DataTable
from bagels.components.indicators import EmptyIndicator
from bagels.config import CONFIG
from bagels.managers.persons import (
    get_persons_with_splits,
)
from bagels.managers.records import (
    get_record_total_split_amount,
    get_records,
)
from bagels.utils.format import format_date_to_readable


class DisplayMode:
    DATE = "d"
    PERSON = "p"


class RecordTableBuilder:
    def rebuild(self, focus=True) -> None:
        if not hasattr(self, "table"):
            return
        table = self.table
        empty_indicator: EmptyIndicator = self.query_one(".empty-indicator")
        self._initialize_table(table)
        records = self._fetch_records()

        match self.displayMode:
            case DisplayMode.PERSON:
                self._build_person_view(table, records)
            case DisplayMode.DATE:
                self._build_date_view(table, records)
            case _:
                pass

        if hasattr(self, "current_row_index"):
            table.move_cursor(row=self.current_row_index)
        empty_indicator.display = not table.rows
        table.display = not not table.rows
        if focus:
            if table.display:
                table.focus()
            else:
                self.focus()

    def _fetch_records(self):
        params = {
            "offset": self.page_parent.filter["offset"],
            "offset_type": self.page_parent.filter["offset_type"],
        }
        if self.page_parent.filter["byAccount"]:
            params["account_id"] = self.page_parent.mode["accountId"]["default_value"]
        if self.FILTERS["enabled"]():
            params["category_piped_names"] = self.FILTERS["category"]()
            params["operator_amount"] = self.FILTERS["amount"]()
            params["label"] = self.FILTERS["label"]()
        return get_records(
            **params,
        )

    def _initialize_table(self, table: DataTable) -> None:
        table.clear()
        table.columns.clear()
        match self.displayMode:
            case DisplayMode.PERSON:
                table.add_columns(
                    " ",
                    "Date",
                    "Record date",
                    "Category",
                    "Amount",
                    "To account",
                    "Label",
                )
            case DisplayMode.DATE:
                table.add_columns(" ", "Category", "Amount", "Label", "Account")

    def _get_label_string(self, text) -> str:
        if self.FILTERS["enabled"]():
            highlight_style = self.get_component_rich_style("label-highlight-match")
            text = Text(text)
            text.highlight_words(
                [self.FILTERS["label"]()],
                style=highlight_style,
                case_sensitive=False,
            )
        return text

    # region Date view
    def _build_date_view(self, table: DataTable, records: list) -> None:
        prev_group = None
        for record in records:
            flow_icon = self._get_flow_icon(len(record.splits) > 0, record.isIncome)

            category_string, amount_string, account_string = self._format_record_fields(
                record, flow_icon
            )

            # Highlight label if filtering
            label_string = self._get_label_string(record.label)

            # Add group header based on filter type
            group_string = None
            match self.page_parent.filter["offset_type"]:
                case "year":
                    # Group by month
                    group_string = record.date.strftime("%B %Y")
                case "month":
                    # Group by week
                    week_start = record.date - timedelta(
                        days=(record.date.weekday() - CONFIG.defaults.first_day_of_week)
                        % 7
                    )
                    week_end = week_start + timedelta(days=6)

                    # Adjust week_start and week_end if they are not in the same month as record.date
                    if week_start.month != record.date.month:
                        week_start = record.date.replace(day=1)
                    if week_end.month != record.date.month:
                        last_day_of_month = (
                            record.date.replace(day=1) + timedelta(days=32)
                        ).replace(day=1) - timedelta(days=1)
                        week_end = last_day_of_month

                    group_string = f"{format_date_to_readable(week_start)} - {format_date_to_readable(week_end)}"
                case "week":
                    # Group by day
                    group_string = format_date_to_readable(record.date)
                case "day":
                    # No grouping
                    pass

            if group_string and prev_group != group_string:
                prev_group = group_string
                self._add_group_header_row(table, group_string)

            # Add main record row
            table.add_row(
                " ",
                category_string,
                amount_string,
                label_string,
                account_string,
                key=f"r-{str(record.id)}",
            )

            # Add split rows if applicable
            if record.splits and self.show_splits:
                self._add_split_rows(table, record, flow_icon)

    def _get_flow_icon(self, recordHasSplits: bool, is_income: bool) -> str:
        if recordHasSplits and not self.show_splits:
            flow_icon_positive = "[green]=[/green]"
            flow_icon_negative = "[red]=[/red]"
        else:
            flow_icon_positive = f"[green]{CONFIG.symbols.amount_positive}[/green]"
            flow_icon_negative = f"[red]{CONFIG.symbols.amount_negative}[/red]"
        return flow_icon_positive if is_income else flow_icon_negative

    def _format_record_fields(self, record, flow_icon: str) -> tuple[str, str]:
        if record.isTransfer:
            from_account = (
                "[italic]" + record.account.name + "[/italic]"
                if record.account.hidden
                else record.account.name
            )
            to_account = (
                "[italic]" + record.transferToAccount.name + "[/italic]"
                if record.transferToAccount.hidden
                else record.transferToAccount.name
            )
            category_string = f"{from_account} → {to_account}"
            amount_string = record.amount
            account_string = "-"
        else:
            color_tag = record.category.color.lower()
            category_string = f"[{color_tag}]{CONFIG.symbols.category_color}[/{color_tag}] {record.category.name}"

            if record.splits and not self.show_splits:
                amount_self = round(
                    record.amount - get_record_total_split_amount(record.id), 2
                )
                amount_string = f"{flow_icon} {amount_self}"
            else:
                amount_string = f"{flow_icon} {record.amount}"

            account_string = record.account.name

        return category_string, amount_string, account_string

    def _add_group_header_row(
        self, table: DataTable, string: str, key: str = None
    ) -> None:
        table.add_row("//", string, "", "", "", style_name="group-header", key=key)

    def _add_split_rows(self, table: DataTable, record, flow_icon: str) -> None:
        color = record.category.color.lower()
        amount_self = round(
            record.amount - get_record_total_split_amount(record.id),
            CONFIG.defaults.round_decimals,
        )
        split_flow_icon = (
            f"[red]{CONFIG.symbols.amount_negative}[/red]"
            if record.isIncome
            else f"[green]{CONFIG.symbols.amount_positive}[/green]"
        )
        line_char = f"[{color}]{CONFIG.symbols.line_char}[/{color}]"
        finish_line_char = f"[{color}]{CONFIG.symbols.finish_line_char}[/{color}]"

        for split in record.splits:
            paid_status_icon = self._get_split_status_icon(split)
            date_string = (
                Text(f"Paid {format_date_to_readable(split.paidDate)}", style="italic")
                if split.paidDate
                else Text("-")
            )

            table.add_row(
                " ",
                f"{line_char} {paid_status_icon} {split.person.name}",
                f"{split_flow_icon} {split.amount}",
                date_string,
                split.account.name if split.account else "-",
                key=f"s-{str(split.id)}",
            )

        # Add net amount row
        table.add_row(
            "",
            f"{finish_line_char} Self total",
            f"= {amount_self}",
            "",
            "",
            style_name="net",
        )

    def _get_split_status_icon(self, split) -> str:
        if split.isPaid:
            return f"[green]{CONFIG.symbols.split_paid}[/green]"
        else:
            return f"[grey]{CONFIG.symbols.split_unpaid}[/grey]"

    # region Person view

    def _fetch_person_records(self) -> list:
        params = {
            "offset": self.page_parent.filter["offset"],
            "offset_type": self.page_parent.filter["offset_type"],
        }
        if self.FILTERS["enabled"]():
            params["category_piped_names"] = self.FILTERS["category"]()
            params["operator_amount"] = self.FILTERS["amount"]()
            params["label"] = self.FILTERS["label"]()
        return get_persons_with_splits(**params)

    def _build_person_view(self, table: DataTable, _) -> None:
        persons = self._fetch_person_records()

        # Display each person and their splits
        for person in persons:
            if person.splits:  # Person has splits for this month
                # Add person header
                self._add_group_header_row(
                    table, person.name, key=f"p-{str(person.id)}"
                )

                # Add splits for this person
                total_unpaid = 0  # Initialize total unpaid amount for this person
                for split in person.splits:
                    record = split.record
                    paid_icon = (
                        f"[green]{CONFIG.symbols.split_paid}[/green]"
                        if split.isPaid
                        else f"[red]{CONFIG.symbols.split_unpaid}[/red]"
                    )
                    date = (
                        format_date_to_readable(split.paidDate)
                        if split.paidDate
                        else "Not paid"
                    )
                    record_date = format_date_to_readable(record.date)
                    category = f"[{record.category.color.lower()}]{CONFIG.symbols.category_color}[/{record.category.color.lower()}] {record.category.name}"

                    # Calculate amount and update total of unpaid amounts
                    if not split.isPaid:
                        split_amount = split.amount
                        if record.isIncome:
                            split_amount = -split_amount  # Negate income amounts
                        total_unpaid += split_amount

                    amount = (
                        f"[red]{CONFIG.symbols.amount_negative}[/red] {split.amount}"
                        if record.isIncome
                        else f"[green]{CONFIG.symbols.amount_positive}[/green] {split.amount}"
                    )
                    account = f"→ {split.account.name}" if split.account else "-"

                    label_string = self._get_label_string(record.label)

                    table.add_row(
                        " ",
                        f"{paid_icon} {date}",
                        record_date,
                        category,
                        amount,
                        account,
                        label_string,
                        key=f"s-{split.id}",
                    )

                # Add total row for this person showing unpaid amount. We reverse the color indicator.
                if total_unpaid == 0:
                    total_display = "0.0"
                elif total_unpaid < 0:
                    total_display = f"[green]{abs(total_unpaid)}[/green]"
                else:
                    total_display = f"[red]{abs(total_unpaid)}[/red]"
                table.add_row(
                    " ",
                    "[bold]Total Unpaid[/bold]",
                    "",
                    "",
                    f"[bold]{total_display}[/bold]",
                    "",
                    key=f"t-{str(person.id)}",
                )



from datetime import datetime

from rich.text import Text
from sqlalchemy import desc, func, select
from sqlalchemy.orm import joinedload, sessionmaker

from bagels.managers.utils import get_start_end_of_period
from bagels.models.category import Category
from bagels.models.database.app import db_engine
from bagels.models.record import Record

Session = sessionmaker(bind=db_engine)


# region Get
def get_categories_count():
    """Count all categories excluding deleted ones."""
    session = Session()
    try:
        stmt = select(Category)
        return len(session.scalars(stmt).all())
    finally:
        session.close()


def get_all_categories_tree() -> list[tuple[Category, Text, int]]:
    """Retrieve all categories in a hierarchical tree format."""
    session = Session()
    try:
        stmt = (
            select(Category)
            .options(joinedload(Category.parentCategory))
            .order_by(Category.id)
            .filter(Category.deletedAt.is_(None))
        )
        categories = session.scalars(stmt).all()

        def build_category_tree(parent_id=None, depth=0):
            result = []
            for category in categories:
                if category.parentCategoryId == parent_id:
                    if depth == 0:
                        node = Text("●", style=category.color)
                    else:
                        node = Text(
                            " " * (depth - 1)
                            + ("└" if is_last(category, parent_id) else "├"),
                            style=category.color,
                        )
                    result.append((category, node, depth))
                    result.extend(build_category_tree(category.id, depth + 1))
            return result

        def is_last(category, parent_id):
            siblings = [cat for cat in categories if cat.parentCategoryId == parent_id]
            return category == siblings[-1]

        return build_category_tree()
    finally:
        session.close()


def get_all_categories_by_freq():
    """Retrieve all categories ordered by the frequency of their usage in records."""
    session = Session()
    try:
        stmt = (
            select(Category, func.count(Category.records).label("record_count"))
            .outerjoin(Category.records)
            .group_by(Category.id)
            .order_by(desc("record_count"))
            .options(joinedload(Category.parentCategory))
            .filter(Category.deletedAt.is_(None))
        )
        return session.execute(stmt).all()
    finally:
        session.close()


def get_category_by_id(category_id):
    """Retrieve a category by its ID."""
    session = Session()
    try:
        stmt = (
            select(Category)
            .filter_by(id=category_id)
            .filter(Category.deletedAt.is_(None))
        )
        return session.scalars(stmt).first()
    finally:
        session.close()


def get_all_categories_records(
    offset: int = 0,
    offset_type: str = "month",
    is_income: bool = True,
    subcategories: bool = False,
    account_id: int = None,
):
    """
    Retrieve all categories with their net income or expenses, sorted by total amount.
    """
    session = Session()
    try:
        start_of_period, end_of_period = get_start_end_of_period(offset, offset_type)

        stmt = select(Record).options(joinedload(Record.category))
        if account_id is not None:
            stmt = stmt.filter(Record.accountId == account_id)
        stmt = stmt.filter(
            Record.date >= start_of_period,
            Record.date < end_of_period,
            Record.isIncome == is_income,
        )

        category_totals = {}
        records = session.scalars(stmt).all()
        for record in records:
            split_total = sum(split.amount for split in record.splits)
            record_amount = record.amount - split_total

            if record.category is None:
                continue

            category_id = record.categoryId
            if not subcategories and record.category.parentCategoryId:
                category_id = record.category.parentCategoryId

            if category_id not in category_totals:
                category_totals[category_id] = 0
            category_totals[category_id] += record_amount

        stmt = (
            select(Category)
            .filter(
                Category.id.in_(category_totals.keys()), Category.deletedAt.is_(None)
            )
            .options(joinedload(Category.parentCategory))
        )
        categories = session.scalars(stmt).all()
        for category in categories:
            category.amount = category_totals[category.id]

        categories = [cat for cat in categories if cat.amount != 0]
        categories.sort(key=lambda cat: cat.amount, reverse=True)

        return categories
    finally:
        session.close()


# region Create
def create_category(data):
    """Create a new category."""
    session = Session()
    try:
        new_category = Category(**data)
        session.add(new_category)
        session.commit()
        session.refresh(new_category)
        session.expunge(new_category)
        return new_category
    finally:
        session.close()


# region Update
def update_category(category_id, data):
    """Update a category by its ID."""
    session = Session()
    try:
        category = session.get(Category, category_id)
        if category:
            for key, value in data.items():
                setattr(category, key, value)
            session.commit()
            session.refresh(category)
            session.expunge(category)
        return category
    finally:
        session.close()


# region Delete
def delete_category(category_id):
    """Delete a category by marking it and its subcategories as deleted."""
    session = Session()
    try:
        category = session.get(Category, category_id)
        if category:
            category.deletedAt = datetime.now()

            # Delete subcategories
            subcategories = (
                session.query(Category).filter_by(parentCategoryId=category_id).all()
            )
            for subcategory in subcategories:
                subcategory.deletedAt = datetime.now()

            session.commit()
            session.refresh(category)
            session.expunge(category)
            return True
        return False
    finally:
        session.close()



from pathlib import Path

from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import MarkdownViewer, Static

from bagels.components.bagel import Bagel


class Welcome(Static):
    can_focus = True

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, id="welcome-container")
        file_path = Path(__file__).parent.parent.parent / "static" / "welcome.md"
        with open(file_path, "r") as file:
            self.welcome_text = file.read()

    def compose(self) -> ComposeResult:
        with Container(classes="text-container"):
            yield MarkdownViewer(self.welcome_text, show_table_of_contents=False)
        yield Bagel()



import sqlite3
from datetime import datetime


def convert_date(date_str):
    """Convert YYYYMMDD to datetime with time set to 12:00 PM"""
    if not date_str:
        return datetime.now().replace(hour=12, minute=0, second=0, microsecond=0)
    try:
        # Convert from YYYYMMDD format
        year = int(str(date_str)[:4])
        month = int(str(date_str)[4:6])
        day = int(str(date_str)[6:8])
        return datetime(year, month, day, 12, 0)
    except Exception:
        return datetime.now().replace(hour=12, minute=0, second=0, microsecond=0)


def ensure_default_category(bagels_cur):
    """Create or get default category for uncategorized transactions."""
    bagels_cur.execute(
        """
        INSERT OR IGNORE INTO category (
            createdAt, updatedAt, name, nature, color, parentCategoryId
        ) VALUES (?, ?, ?, ?, ?, ?)
    """,
        (datetime.now(), datetime.now(), "Uncategorized", "WANT", "#808080", None),
    )

    default_category = bagels_cur.execute("""
        SELECT id FROM category 
        WHERE name = 'Uncategorized' AND parentCategoryId IS NULL
        AND deletedAt IS NULL
    """).fetchone()

    return default_category[0]


class BudgetToBagelsMigration:
    def __init__(self, budget_db_path: str, bagels_db_path: str):
        self.budget_conn = sqlite3.connect(budget_db_path)
        self.bagels_conn = sqlite3.connect(bagels_db_path)
        self.budget_cur = self.budget_conn.cursor()
        self.bagels_cur = self.bagels_conn.cursor()
        self.category_map = {}
        self.account_map = {}
        self.default_category_id = None

    def migrate_accounts(self):
        accounts = self.budget_cur.execute("""
            SELECT id, name, balance_current, official_name, offbudget, closed 
            FROM accounts WHERE tombstone = 0
        """).fetchall()

        for acc_id, name, balance, official_name, offbudget, closed in accounts:
            # Convert cents to dollars
            balance_float = float(balance or 0) / 100
            description = official_name if official_name is not None else ""

            self.bagels_cur.execute(
                """
                INSERT INTO account (createdAt, updatedAt, name, description, 
                                   beginningBalance, hidden)
                VALUES (?, ?, ?, ?, ?, ?)
            """,
                (
                    datetime.now(),
                    datetime.now(),
                    name or "",
                    description,
                    balance_float,
                    bool(offbudget or closed),
                ),
            )

            self.account_map[acc_id] = self.bagels_cur.lastrowid

    def migrate_categories(self):
        self.default_category_id = ensure_default_category(self.bagels_cur)

        groups = self.budget_cur.execute("""
            SELECT id, name, is_income 
            FROM category_groups WHERE tombstone = 0
        """).fetchall()

        for group_id, name, is_income in groups:
            self.bagels_cur.execute(
                """
                INSERT INTO category (createdAt, updatedAt, name, nature, color,
                                    parentCategoryId)
                VALUES (?, ?, ?, ?, ?, ?)
            """,
                (
                    datetime.now(),
                    datetime.now(),
                    name or "",
                    "MUST" if is_income else "WANT",
                    "#808080",
                    None,
                ),
            )
            parent_id = self.bagels_cur.lastrowid

            categories = self.budget_cur.execute(
                """
                SELECT id, name, is_income 
                FROM categories 
                WHERE cat_group = ? AND tombstone = 0
            """,
                (group_id,),
            ).fetchall()

            for cat_id, cat_name, cat_is_income in categories:
                self.bagels_cur.execute(
                    """
                    INSERT INTO category (createdAt, updatedAt, name, nature,
                                        color, parentCategoryId)
                    VALUES (?, ?, ?, ?, ?, ?)
                """,
                    (
                        datetime.now(),
                        datetime.now(),
                        cat_name or "",
                        "MUST" if cat_is_income else "WANT",
                        "#808080",
                        parent_id,
                    ),
                )
                self.category_map[cat_id] = self.bagels_cur.lastrowid

    def migrate_transactions(self):
        transactions = self.budget_cur.execute("""
            SELECT id, account, category, amount, date, 
                   starting_balance_flag, transfer_id, is_parent
            FROM v_transactions_internal 
            WHERE tombstone = 0 AND is_child = 0
            AND (transfer_id IS NULL OR amount >= 0)
        """).fetchall()

        for (
            trans_id,
            account_id,
            category_id,
            amount,
            date,
            is_starting_balance,
            transfer_id,
            is_parent,
        ) in transactions:
            if account_id not in self.account_map:
                continue

            transaction_date = convert_date(date)

            amount_float = abs(float(amount or 0) / 100)
            if amount_float == 0:
                continue

            account_bagels_id = self.account_map[account_id]
            category_bagels_id = self.category_map.get(
                category_id, self.default_category_id
            )

            category_info = self.budget_cur.execute(
                """
                SELECT is_income FROM categories WHERE id = ?
            """,
                (category_id,),
            ).fetchone()
            is_income = category_info[0] if category_info else 0

            transfer_to_account_id = None
            is_transfer = bool(transfer_id)
            if transfer_id:
                transfer_account = self.budget_cur.execute(
                    """
                    SELECT account FROM v_transactions_internal WHERE id = ?
                """,
                    (transfer_id,),
                ).fetchone()
                if transfer_account and transfer_account[0] in self.account_map:
                    transfer_to_account_id = self.account_map[transfer_account[0]]

            self.bagels_cur.execute(
                """
                INSERT INTO record (
                    createdAt, updatedAt, label, amount, date,
                    accountId, categoryId, isIncome, isTransfer,
                    transferToAccountId, isInProgress
                ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
            """,
                (
                    datetime.now(),
                    datetime.now(),
                    "Imported transaction",
                    amount_float,
                    transaction_date,
                    account_bagels_id,
                    category_bagels_id,
                    bool(is_income),
                    is_transfer,
                    transfer_to_account_id,
                    False,
                ),
            )

    def verify_and_fix_categories(self):
        """Verify all records have valid categories and fix any issues"""
        self.bagels_cur.execute(
            """
            UPDATE record 
            SET categoryId = ? 
            WHERE categoryId IS NULL OR 
                  categoryId NOT IN (SELECT id FROM category)
        """,
            (self.default_category_id,),
        )

        null_categories = self.bagels_cur.execute("""
            SELECT COUNT(*) 
            FROM record 
            WHERE categoryId IS NULL
        """).fetchone()[0]

        if null_categories > 0:
            raise Exception(
                f"Found {null_categories} records with NULL categories after fix"
            )

    def migrate(self):
        self.bagels_conn.execute("BEGIN TRANSACTION")
        try:
            self.migrate_accounts()
            self.migrate_categories()
            self.migrate_transactions()
            self.verify_and_fix_categories()
            self.bagels_conn.commit()
            print("Migration completed successfully!")
        except Exception as e:
            self.bagels_conn.rollback()
            raise e
        finally:
            self.budget_conn.close()
            self.bagels_conn.close()


if __name__ == "__main__":
    migrator = BudgetToBagelsMigration("db.sqlite", "bagels.db")
    migrator.migrate()



from textual.app import ComposeResult
from textual.widgets import Label, Static

from bagels.bagel import get_string, phi_spacing, render_frame, theta_spacing


class Bagel(Static):
    A = B = 1

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, classes="bagel-container")

    def on_mount(self) -> None:
        if not self.app.is_testing:
            self.set_interval(1 / 10, self.update_bagel)
        else:
            self.update_bagel()

    def update_bagel(self) -> None:
        bagel = self.query_one("#bagel")
        self.A += theta_spacing
        self.B += phi_spacing
        bagel.update(get_string(render_frame(self.A, 1)))

    def compose(self) -> ComposeResult:
        yield Label(id="bagel")



from abc import ABC, abstractmethod
from datetime import datetime
from functools import lru_cache

import numpy as np

from bagels.components.tplot.plot import Plot
from bagels.config import CONFIG
from bagels.managers.records import (
    get_daily_balance,
    get_spending,
    get_spending_trend,
)
from bagels.managers.utils import get_income_to_use


class BasePlot(ABC):
    name: str = "Base Plot"
    supports_cross_periods: bool = False

    def __init__(self, app):
        self.app = app

    @abstractmethod
    def get_data(
        self, start_of_period: datetime, end_of_period: datetime
    ) -> list[float]:
        """Return a list of data points"""
        pass

    @abstractmethod
    def plot(
        self,
        plt: Plot,
        start_of_period: datetime,
        end_of_period: datetime,
        offset: int,
        data: list[float],
        dates: list[str],
        get_theme_color,
    ) -> None:
        """Additional operations on the plotext object."""
        pass


class SpendingPlot(BasePlot):
    name: str = "Spending"

    @lru_cache
    def get_data(self, start_of_period, end_of_period):
        return get_spending(start_of_period, end_of_period)

    def plot(
        self,
        plt: Plot,
        start_of_period: datetime,
        end_of_period: datetime,
        offset: int,
        data: list[float],
        dates: list[str],
        get_theme_color,
    ) -> None:
        if min(data) >= 0:
            plt.ylim(lower=0)


class SpendingTrajectoryPlot(BasePlot):
    name: str = "Spending Trajectory"

    @lru_cache
    def get_data(self, start_of_period, end_of_period):
        return get_spending_trend(start_of_period, end_of_period)

    def plot(
        self,
        plt: Plot,
        start_of_period: datetime,
        end_of_period: datetime,
        offset: int,
        data: list[float],
        dates: list[str],
        get_theme_color,
    ) -> None:
        # --------- Limit computation -------- #

        limit = get_income_to_use(offset)

        plt.ylim(upper=limit, lower=0)

        if len(data) == len(dates):
            return  # don't have to show regression prediction trend

        # ---------- Prediction line --------- #

        # Estimate data trend by creating an array of length len(dates) - len(data), filled with values from linear regression.
        # Plot the data by using reversed dates to put at the right hand side of the plot
        if len(data) >= 2:
            x = np.arange(len(data))
            coefficients = np.polyfit(x, data, 1)
            trend = np.poly1d(coefficients)

            # Generate prediction points for the remaining days
            remaining_days = len(dates) - len(data)
            if remaining_days > 0:
                prediction_x = np.arange(len(data), len(dates))
                prediction_y = trend(prediction_x)
                prediction_data = data + prediction_y.tolist()

                plt.plot(
                    dates,
                    prediction_data,
                    marker=CONFIG.defaults.plot_marker,
                    color=get_theme_color("secondary"),
                )

        # ----- Period spending separator ---- #

        period_spending = max(data)

        total_days = (end_of_period - start_of_period).days + 1

        plt.plot(
            dates,
            [period_spending] * total_days,
            marker=CONFIG.defaults.plot_marker,
            color=get_theme_color("panel"),
        )


class BalancePlot(BasePlot):
    name: str = "Balance"
    supports_cross_periods = True

    @lru_cache
    def get_data(self, start_of_period, end_of_period):
        return get_daily_balance(start_of_period, end_of_period)

    def plot(
        self,
        plt: Plot,
        start_of_period: datetime,
        end_of_period: datetime,
        offset: int,
        data: list[float],
        dates: list[str],
        get_theme_color,
    ) -> None:
        pass



from dataclasses import dataclass

from sqlalchemy import and_, func, select
from sqlalchemy.orm import contains_eager, sessionmaker

from bagels.managers.utils import get_operator_amount, get_start_end_of_period
from bagels.models.category import Category
from bagels.models.database.app import db_engine
from bagels.models.person import Person
from bagels.models.record import Record
from bagels.models.split import Split

Session = sessionmaker(bind=db_engine)


# region Create


def create_person(data):
    """Create a new person entry in the database."""
    session = Session()
    try:
        new_person = Person(**data)
        session.add(new_person)
        session.commit()
        session.refresh(new_person)
        session.expunge(new_person)
        return new_person
    finally:
        session.close()


# region Read


def get_all_persons() -> list[Person]:
    """Retrieve all undeleted persons from the database."""
    session = Session()
    try:
        return session.scalars(select(Person).where(Person.deletedAt.is_(None))).all()
    finally:
        session.close()


def get_person_by_id(person_id) -> Person:
    """Retrieve a person by their ID."""
    session = Session()
    try:
        return session.get(Person, person_id)
    finally:
        session.close()


def get_persons_with_splits(
    offset: int = 0,
    offset_type: str = "month",
    category_piped_names: str = None,
    operator_amount: str = None,
    label: str = None,
):
    """Get all persons with their splits for the specified period."""
    session = Session()
    try:
        start_of_period, end_of_period = get_start_end_of_period(offset, offset_type)

        # Build the base query
        stmt = (
            select(Person)
            .options(
                contains_eager(Person.splits)
                .contains_eager(Split.record)
                .contains_eager(Record.category),
                contains_eager(Person.splits).contains_eager(Split.account),
            )
            .join(Person.splits)
            .join(Split.record)
            .outerjoin(Record.category)
            .outerjoin(Split.account)
        )

        # Apply date filter
        stmt = stmt.filter(
            and_(Record.date >= start_of_period, Record.date < end_of_period)
        )

        # Apply category filter
        if category_piped_names not in [None, ""]:
            category_names = category_piped_names.split("|")
            stmt = stmt.filter(Category.name.in_(category_names))

        # Apply amount filter
        if operator_amount not in [None, ""]:
            operator, amount = get_operator_amount(operator_amount)
            if operator and amount:
                stmt = stmt.filter(Split.amount.op(operator)(amount))

        # Apply label filter
        if label not in [None, ""]:
            stmt = stmt.filter(Record.label.ilike(f"%{label}%"))

        # Apply ordering and distinct
        stmt = stmt.order_by(Record.date.asc()).distinct()

        result = session.scalars(stmt)
        return result.unique().all()
    finally:
        session.close()


@dataclass
class PersonWithDue:
    person: Person
    due: float


def get_persons_with_net_due() -> list[Person]:
    """Retrieve all persons with their net due amount."""
    session = Session()
    try:
        stmt = (
            select(
                Person,
                (
                    func.coalesce(
                        select(func.sum(Split.amount))
                        .select_from(Split)
                        .join(Record)
                        .where(
                            Split.personId == Person.id,
                            Record.isIncome == False,  # noqa: E712
                            Split.isPaid == False,  # noqa: E712
                        )
                        .correlate(Person)
                        .scalar_subquery(),
                        0,
                    )
                    - func.coalesce(
                        select(func.sum(Split.amount))
                        .select_from(Split)
                        .join(Record)
                        .where(
                            Split.personId == Person.id,
                            Record.isIncome == True,  # noqa: E712
                            Split.isPaid == False,  # noqa: E712
                        )
                        .correlate(Person)
                        .scalar_subquery(),
                        0,
                    )
                ).label("due"),
            )
            .select_from(Person)
            .where(Person.deletedAt.is_(None))
            .order_by(Person.name)
        )

        result = session.execute(stmt).all()
        persons_with_due = []
        for person, due in result:
            person.due = due
            persons_with_due.append(person)
        return persons_with_due
    finally:
        session.close()


# region Update


def update_person(person_id, data) -> Person:
    """Update a person's information by their ID."""
    session = Session()
    try:
        person = session.get(Person, person_id)
        if person:
            for key, value in data.items():
                setattr(person, key, value)
            session.commit()
            session.refresh(person)
            session.expunge(person)
        return person
    finally:
        session.close()


# region Delete


def delete_person(person_id) -> bool:
    """Delete a person - if they have splits, soft delete by setting deletedAt, otherwise hard delete."""
    session = Session()
    try:
        person = session.get(Person, person_id)
        if person:
            # Check if person has any splits
            has_splits = (
                session.query(Split).filter(Split.personId == person_id).first()
                is not None
            )

            if has_splits:
                # Soft delete if person has splits
                person.deletedAt = func.now()
            else:
                # Hard delete if person has no splits
                session.delete(person)

            session.commit()
            return True
        return False
    finally:
        session.close()



from textual.app import ComposeResult
from textual.binding import Binding
from textual.widgets import Static

from bagels.components.datatable import DataTable
from bagels.config import CONFIG
from bagels.forms.category_form import CategoryForm
from bagels.managers.categories import (
    create_category,
    delete_category,
    get_all_categories_tree,
    get_category_by_id,
    update_category,
)
from bagels.modals.confirmation import ConfirmationModal
from bagels.modals.input import InputModal


class Categories(Static):
    can_focus = True

    COLUMNS = ("", "Name", "Nature")

    BINDINGS = [
        Binding(
            CONFIG.hotkeys.categories.browse_defaults,
            "browse_defaults",
            "Browse",
            False,
        ),
        Binding(CONFIG.hotkeys.new, "new_category", "Add"),
        Binding(
            CONFIG.hotkeys.categories.new_subcategory,
            "new_subcategory",
            "Add Subcategory",
        ),
        Binding(CONFIG.hotkeys.edit, "edit_category", "Edit"),
        Binding(CONFIG.hotkeys.delete, "delete_category", "Delete"),
    ]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="categories-container", classes="module-container"
        )
        super().__setattr__("border_title", "Categories")

    # --------------- Hooks -------------- #

    def on_mount(self) -> None:
        self.rebuild()

    def on_data_table_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        if event.row_key:
            self.current_row = event.row_key.value

    # region Builders
    # ------------- Builders ------------- #

    def rebuild(self) -> None:
        table: DataTable = self.query_one("#categories-table")
        # empty_indicator: Static = self.query_one(".empty-indicator")

        table.clear()
        if not table.columns:
            table.add_columns(*self.COLUMNS)

        categories = get_all_categories_tree()
        if categories:
            for category, node, depth in categories:
                char = " "
                nature = category.nature.value
                if depth == 0:
                    char = ""
                table.add_row(
                    node,
                    char + category.name,
                    nature,
                    key=category.id,
                )
            table.zebra_stripes = True
        else:
            self.current_row = None

        # empty_indicator.display = not categories

    # region Helpers
    # -------------- Helpers ------------- #

    def _notify_no_categories(self) -> None:
        self.app.notify(
            title="Error",
            message="Category must be selected for this action.",
            severity="error",
            timeout=2,
        )

    # region callbacks
    # ------------- Callbacks ------------ #

    def action_new_category(self) -> None:
        def check_result(result) -> None:
            if result:
                try:
                    create_category(result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message="Category created",
                        severity="information",
                        timeout=3,
                    )
                    self.rebuild()

        self.app.push_screen(
            InputModal("New Category", CategoryForm().get_form()), callback=check_result
        )

    def action_new_subcategory(self) -> None:
        if not self.current_row:
            self._notify_no_categories()
            return

        def check_result(result) -> None:
            if result:
                try:
                    create_category(result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message="Subcategory created",
                        severity="information",
                        timeout=3,
                    )
                    self.rebuild()

        parent_category_id = self.current_row
        subcategory_form = CategoryForm().get_subcategory_form(parent_category_id)
        parent_category = get_category_by_id(parent_category_id)
        self.app.push_screen(
            InputModal(f"New Subcategory of {parent_category.name}", subcategory_form),
            callback=check_result,
        )

    def action_delete_category(self) -> None:
        if not self.current_row:
            self._notify_no_categories()
            return

        def check_delete(result) -> None:
            if result:
                try:
                    delete_category(self.current_row)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                self.rebuild()

        category = get_category_by_id(self.current_row)

        self.app.push_screen(
            ConfirmationModal(
                f"Are you sure you want to delete category '[{category.color}]●[/{category.color}] {category.name}'?"
            ),
            check_delete,
        )

    def action_edit_category(self) -> None:
        if not self.current_row:
            self._notify_no_categories()
            return

        def check_result(result) -> None:
            if result:
                try:
                    update_category(self.current_row, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Category {result['name']} updated",
                        severity="information",
                        timeout=3,
                    )
                    self.rebuild()

        filled_form = CategoryForm().get_filled_form(self.current_row)
        self.app.push_screen(
            InputModal("Edit Category", filled_form), callback=check_result
        )

    # region View
    # --------------- View --------------- #
    def compose(self) -> ComposeResult:
        yield DataTable(
            id="categories-table",
            cursor_type="row",
            cursor_foreground_priority=True,
        )
        # yield EmptyIndicator("No categories")



import os
import platform
import subprocess
import warnings
from typing import Any, Literal

import yaml
from pydantic import BaseModel, Field, ValidationError

from bagels.locations import config_file


class Defaults(BaseModel):
    period: Literal["day", "week", "month", "year"] = "week"
    first_day_of_week: int = Field(ge=0, le=6, default=6)
    date_format: str = "%d/%m"
    round_decimals: int = 2
    plot_marker: Literal["braille", "fhd", "hd", "dot"] = "braille"


class DatemodeHotkeys(BaseModel):
    go_to_day: str = "g"


class HomeHotkeys(BaseModel):
    cycle_tabs: str = "c"
    budgets: str = "b"
    new_transfer: str = "t"
    toggle_splits: str = "s"
    display_by_date: str = "q"
    display_by_person: str = "w"
    advance_filter: str = "f"
    cycle_offset_type: str = "."
    toggle_income_mode: str = "/"
    select_prev_account: str = "["
    select_next_account: str = "]"
    toggle_use_account: str = "\\"
    datemode: DatemodeHotkeys = DatemodeHotkeys()


class RecordModalHotkeys(BaseModel):
    new_split: str = "ctrl+a"
    new_paid_split: str = "ctrl+s"
    delete_last_split: str = "ctrl+d"


class CategoriesHotkeys(BaseModel):
    new_subcategory: str = "s"
    browse_defaults: str = "b"


class Hotkeys(BaseModel):
    new: str = "a"
    delete: str = "d"
    edit: str = "e"
    toggle_jump_mode: str = "v"
    home: HomeHotkeys = HomeHotkeys()
    record_modal: RecordModalHotkeys = RecordModalHotkeys()
    categories: CategoriesHotkeys = CategoriesHotkeys()


class Symbols(BaseModel):
    line_char: str = "│"
    finish_line_char: str = "╰"
    split_paid: str = "✓"
    split_unpaid: str = "⨯"
    category_color: str = "●"
    amount_positive: str = "+"
    amount_negative: str = "-"


class BudgetingStates(BaseModel):
    # ---------- Income policies --------- #
    income_assess_metric: Literal["periodIncome", "fallback"] = (
        "periodIncome"  # use the current period's income, or if less than threshold, use the past period's income
    )
    income_assess_threshold: float = (
        100  # if income less than this, we assume has no income
    )
    income_assess_fallback: float = (
        3500  # if income less than threshold, we use this as the income
    )
    # -------- Savings budgetting -------- #
    savings_assess_metric: Literal["percentagePeriodIncome", "setAmount"] = (
        "percentagePeriodIncome"
    )
    savings_percentage: float = (
        0.2  # used only if savings_assess_metric is percentagePeriodIncome
    )
    savings_amount: float = 0  # used only if savings_assess_metric is setAmount
    # ---------- MNW budgetting ---------- #
    wants_spending_assess_metric: Literal["percentageQuota", "setAmount"] = (
        "percentageQuota"  # percentage of all expenses
    )
    wants_spending_percentage: float = (
        0.2  # used only if wants_spending_assess_metric is setPercentage
    )
    wants_spending_amount: float = (
        0  # used only if wants_spending_assess_metric is setAmount
    )


class State(BaseModel):
    theme: str = "dark"
    check_for_updates: bool = True
    footer_visibility: bool = True
    budgeting: BudgetingStates = BudgetingStates()


class Config(BaseModel):
    hotkeys: Hotkeys = Hotkeys()
    symbols: Symbols = Symbols()
    defaults: Defaults = Defaults()
    state: State = State()

    def __init__(self, **data):
        try:
            config_data = self._load_yaml_config()
            merged_data = {**self.model_dump(), **config_data, **data}
            super().__init__(**merged_data)
            self.ensure_yaml_fields()
        except ValidationError as e:
            error_messages = []
            for error in e.errors():
                field = error["loc"]
                field_path = ".".join(str(x) for x in field)
                input_value = error.get("input")
                allowed_values = None

                # Extract allowed values for literal errors
                if error["type"] == "literal_error":
                    # Parse the error message to extract allowed values
                    msg = error["msg"]
                    allowed_list = msg.split("'")[1::2]  # Extract values between quotes
                    allowed_values = " or ".join(f"'{v}'" for v in allowed_list)

                message = f"Invalid configuration in field '{field_path}'"
                if input_value is not None:
                    message += f"\nCurrent value: '{input_value}'"
                if allowed_values:
                    message += f"\nAllowed values: {allowed_values}"

                error_messages.append(message)

            raise ConfigurationError("\n\n".join(error_messages))

    def _load_yaml_config(self) -> dict[str, Any]:
        config_path = config_file()
        if not config_path.is_file():
            return {}

        try:
            with open(config_path, "r") as f:
                config = yaml.safe_load(f)
                return config if isinstance(config, dict) else {}
        except Exception as e:
            warnings.warn(f"Error loading config file: {e}")
            return {}

    def ensure_yaml_fields(self):
        try:
            with open(config_file(), "r") as f:
                config = yaml.safe_load(f) or {}
        except FileNotFoundError:
            config = {}

        def update_config(default, current):
            for key, value in default.items():
                if isinstance(value, dict):
                    current[key] = update_config(value, current.get(key, {}))
                elif key not in current:
                    current[key] = value
            return current

        default_config = self.model_dump()
        config = update_config(default_config, config)

        with open(config_file(), "w") as f:
            yaml.dump(config, f, default_flow_style=False)

    @classmethod
    def get_default(cls):
        return cls(
            hotkeys=Hotkeys(), symbols=Symbols(), defaults=Defaults(), state=State()
        )


class ConfigurationError(Exception):
    """Custom exception for configuration errors"""

    pass


CONFIG = None


def open_config_file():
    """Open the config file with the default application."""
    config_path = config_file()
    if platform.system() == "Darwin":  # macOS
        subprocess.run(["open", config_path])
    elif platform.system() == "Windows":  # Windows
        os.startfile(config_path)
    else:  # Linux
        subprocess.run(["xdg-open", config_path])


def load_config():
    f = config_file()
    if not f.exists():
        try:
            f.touch()
            with open(f, "w") as f:
                yaml.dump(Config.get_default().model_dump(), f)
        except OSError:
            pass

    global CONFIG
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            CONFIG = Config()  # ignore warnings about empty env file
    except ConfigurationError as e:
        print("\nConfiguration Error:")
        print("==================")
        print(f"{e}\n")
        print("Would you like to open the config file to fix this? (y/n)")

        try:
            response = input().lower()
            if response.startswith("y"):
                open_config_file()
                print(
                    "\nOpened config file. Please fix the error and restart the application."
                )
            else:
                print(
                    "\nPlease update your config.yaml file with valid values and try again."
                )
        except KeyboardInterrupt:
            print("\nExiting...")

        raise SystemExit(1)


def write_state(key: str, value: Any) -> None:
    """Write a state value to the config.yaml file, supporting nested keys with dot operator."""
    try:
        with open(config_file(), "r") as f:
            config = yaml.safe_load(f) or {}
    except FileNotFoundError:
        config = {}

    keys = key.split(".")
    d = config.setdefault("state", {})
    for k in keys[:-1]:
        d = d.setdefault(k, {})
    d[keys[-1]] = value

    with open(config_file(), "w") as f:
        yaml.dump(config, f, default_flow_style=False)

    # update the global config object
    global CONFIG
    d = CONFIG.state
    for k in keys[:-1]:
        d = getattr(d, k)
    setattr(d, keys[-1], value)



import re
from datetime import datetime, timedelta
from functools import lru_cache

from sqlalchemy.orm import sessionmaker
from textual.widget import Widget

from bagels.config import CONFIG
from bagels.models.category import Category
from bagels.models.database.app import db_engine
from bagels.models.record import Record

Session = sessionmaker(bind=db_engine)

# --------------- query -------------- #


def try_method_query_one(widget: Widget, query: str, method: str, params):
    try:
        widget = widget.query_one(query)
        getattr(widget, method)(*params)
    except Exception as e:
        print(e)
        return


# region period
# -------------- period -------------- #
def _get_start_end_of_year(offset: int = 0):
    now = datetime.now()
    target_year = now.year + offset
    start_of_year = datetime(target_year, 1, 1, 0, 0, 0)
    end_of_year = datetime(target_year, 12, 31, 23, 59, 59)
    return start_of_year, end_of_year


def _get_start_end_of_month(offset: int = 0):
    now = datetime.now()
    # Calculate target month and year
    target_month = now.month + offset
    target_year = now.year + (target_month - 1) // 12
    target_month = ((target_month - 1) % 12) + 1

    # Calculate next month and year for end date
    next_month = target_month + 1
    next_year = target_year + (next_month - 1) // 12
    next_month = ((next_month - 1) % 12) + 1

    start_of_month = datetime(target_year, target_month, 1, 0, 0, 0)
    end_of_month = datetime(next_year, next_month, 1, 0, 0, 0) - timedelta(seconds=1)

    return start_of_month, end_of_month


def _get_start_end_of_week(offset: int = 0):
    now = datetime.now()
    # Apply offset in weeks
    now = now + timedelta(weeks=offset)
    first_day_of_week = CONFIG.defaults.first_day_of_week
    start_of_week = (
        now - timedelta(days=(now.weekday() - first_day_of_week) % 7)
    ).replace(hour=0, minute=0, second=0, microsecond=0)
    end_of_week = (start_of_week + timedelta(days=6)).replace(
        hour=23, minute=59, second=59
    )
    return start_of_week, end_of_week


def _get_start_end_of_day(offset: int = 0):
    now = datetime.now()
    # Apply offset in days
    now = now + timedelta(days=offset)
    start_of_day = now.replace(hour=0, minute=0, second=0, microsecond=0)
    end_of_day = now.replace(hour=23, minute=59, second=59)
    return start_of_day, end_of_day


def get_start_end_of_period(offset: int = 0, offset_type: str = "month"):
    match offset_type:
        case "year":
            return _get_start_end_of_year(offset)
        case "month":
            return _get_start_end_of_month(offset)
        case "week":
            return _get_start_end_of_week(offset)
        case "day":
            return _get_start_end_of_day(offset)


# region figure
# -------------- figure -------------- #


def get_period_figures(
    accountId=None,
    offset_type=None,
    offset=None,
    isIncome=None,
    nature=None,
    session=None,
):
    """Returns the income / expense for a given period.

    Rules:
    - Filter applies to records only by the "date" column, and splits should always be considered by their associated record.
    - Income and expenses should be calculated from record less their splits, regardless of the account of the split.
    - Transfers are not income or expenses, but should be included in the net total.

    Args:
        accountId (int): The ID of the account to filter by. (Optional)
        offset_type (str): The type of period to filter by.
        offset (int): The offset from the current period.
        isIncome (bool): Whether to filter by income or expense.
        nature (Nature): Filter by category nature (Want/Need/Must). (Optional)
        session (Session, optional): SQLAlchemy session to use. If None, creates a new session.
    """
    if session is None:
        session = Session()
        should_close = True
    else:
        should_close = False

    try:
        query = session.query(Record)

        # Filter by account if specified
        if accountId is not None:
            query = query.filter(Record.accountId == accountId)

        # Filter by date period if specified
        if offset_type is not None and offset is not None:
            start_of_period, end_of_period = get_start_end_of_period(
                offset, offset_type
            )
            query = query.filter(
                Record.date >= start_of_period, Record.date < end_of_period
            )

        # Filter by category nature if specified
        if nature is not None:
            query = query.join(Record.category).filter(Category.nature == nature)

        # Calculate net amount
        total = 0
        records = query.all()

        for record in records:
            # Skip records that are marked as transfers if filtering by income type
            if isIncome is not None and record.isTransfer:
                continue

            # Skip records that don't match income filter if specified
            if isIncome is not None and record.isIncome != isIncome:
                continue

            # Calculate amount after subtracting splits
            split_total = sum(split.amount for split in record.splits)
            record_amount = record.amount - split_total

            if not record.isTransfer:
                # For regular income/expenses
                if record.isIncome:
                    total += record_amount
                else:
                    total -= record_amount

        return abs(round(total, CONFIG.defaults.round_decimals))
    finally:
        if should_close:
            session.close()


# region average
# -------------- average ------------- #


def _get_days_in_period(offset: int = 0, offset_type: str = "month"):
    start_of_period, end_of_period = get_start_end_of_period(offset, offset_type)
    days = (end_of_period - start_of_period).days + 1
    return days


def get_period_average(net: int = 0, offset: int = 0, offset_type: str = "month"):
    days = _get_days_in_period(offset, offset_type)
    return round(net / days, CONFIG.defaults.round_decimals)


# region filter process
# ------------ filter process ------------ #


def get_operator_amount(operator_amount: str = None):
    # operators can be >=, >, =, <=, <
    # first validate the string to have one of operators and a number.
    # then split the string to get the operator and the number.
    # then return the operator and amount.
    if re.match(r"^(>=|>|=|<=|<)?\d+(\.\d+)?$", operator_amount):
        if operator_amount[0].isdigit():
            operator, amount = "=", operator_amount
        elif operator_amount[1].isdigit():
            operator, amount = operator_amount[:1], operator_amount[1:]
        else:
            operator, amount = operator_amount[:2], operator_amount[2:]
        amount = float(amount)
        return operator, amount
    else:
        return None, None


# region Budgeting
# ------------- budgeting ------------ #


def get_income_to_use(offset: int):
    metric = CONFIG.state.budgeting.income_assess_metric  # use number if provided
    threshold = CONFIG.state.budgeting.income_assess_threshold
    fallback = CONFIG.state.budgeting.income_assess_fallback

    limit = 0
    if metric == "periodIncome":
        this_month_income = get_period_figures(
            isIncome=True, offset=offset, offset_type="month"
        )
        if this_month_income > threshold:
            limit = this_month_income
        else:
            limit = get_period_figures(
                isIncome=True, offset=offset - 1, offset_type="month"
            )

    if limit < fallback:
        limit = fallback

    return limit


def dynamic_cache(func, *args, **kwargs):
    # Create a cached version of the function
    cached_func = lru_cache()(func)
    # Call the cached version with the provided arguments
    result = cached_func(*args, **kwargs)
    return result



"""Provides a widget for creating and displaying a Plotext plot."""

from __future__ import annotations

from typing import Literal

from plotext._dict import themes as _themes
from rich.text import Text
from textual.app import RenderResult
from textual.color import Color
from textual.reactive import reactive
from textual.widget import Widget

from .plot import Plot, PlotextThemeName, _rgbify_theme, _sequence


class PlotextPlot(Widget):
    """A Plotext plot display widget."""

    DEFAULT_CSS = """
    PlotextPlot {
        width: 1fr;
        height: 1fr;
    }
    """

    theme: reactive[Literal["auto"] | PlotextThemeName] = reactive("auto")
    """The theme to use for the plot.

    If set to `"auto"` the theme will be dynamically generated based on the
    current theme of the Textual app.

    Note that this will just alter the background and foreground colors of the
    plot to match the current Textual theme, not the colors of the plotted graph.

    If set to a specific Plotext theme name, that theme will be used.
    """

    def __init__(
        self,
        *,
        name: str | None = None,
        id: str | None = None,  # pylint:disable=redefined-builtin
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Initialise the Plotext plot widget.

        Args:
            name: The name of the Plotext plot widget.
            id: The ID of the Plotext plot widget in the DOM.
            classes: The CSS classes of the Plotext plot widget.
            disabled: Whether the Plotext plot widget is disabled or not.
        """
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self._plot = Plot()

    def on_mount(self) -> None:
        """Set up the plot."""
        self.app.theme_changed_signal2.subscribe(
            self, lambda theme: self._register_theme(self.app.theme, theme)
        )
        self._register_theme(self.app.theme, self.app.app_theme)

    @property
    def plt(self) -> Plot:
        """The Plotext plotting object.

        Whereas normally Plotext-using code would so something like this:

        ```python
        import plotext as plt

        ...

        plt.some_call()
        plt.some_other_call()
        ```

        use this `plt` property instead of importing `plotext`.
        """
        return self._plot

    def render(self) -> RenderResult:
        """Render the plot.

        Returns:
            The renderable for displaying the plot.
        """
        self._plot.plotsize(self.size.width, self.size.height)
        # This is a belt-and-braces setting of the size of the plot.
        # Internally plotsize calls _set_plot, and as best as I can figure
        # out, what I'm doing here *should* be a no-op (or rather a repeat
        # of what I've just done). And yet... the resizing of plots just
        # doesn't work right without this. This *might* be some
        # as-yet-undiscovered side-effect of dumpster-diving for the figure
        # class.
        #
        # https://github.com/Textualize/textual-plotext/issues/5
        self._plot._set_size(self.size.width, self.size.height)
        # If the theme is set to "auto" ensure the theme is registered with Plotext.
        plotext_theme_name = self._get_plotext_theme_name(self.app.theme)
        self._plot.theme(plotext_theme_name)
        return Text.from_ansi(self._plot.build())

    def _register_theme(self, theme_name: str, app_theme_name: str) -> None:
        """Register the theme with Plotext if necessary.

        Args:
            theme_name: The name of the theme from the Textual app.
              This will only be used if the `theme` reactive of this plot
              is set to `"auto"`.
        """
        # If we're not using auto theme, we don't need to register
        # the Textual app theme with Plotext.
        if self.theme != "auto":
            return

        self.app.log(f"Registering theme {app_theme_name} with Plotext.")
        # I don't know the significance of the plotext theme name, but without it doesn't work so I'm just going to keep it
        plotext_theme_name = self._get_plotext_theme_name(theme_name)
        app_theme_variables = self.app.themes[app_theme_name]
        default_theme_variables = self.app.theme_variables
        _themes[plotext_theme_name] = _rgbify_theme(
            Color.parse(
                app_theme_variables.background
                or default_theme_variables.get("background")
            ).rgb,
            Color.parse(
                app_theme_variables.background
                or default_theme_variables.get("background")
            ).rgb,
            Color.parse(
                app_theme_variables.foreground
                or default_theme_variables.get("foreground")
            ).rgb,
            "default",
            _sequence,
        )
        self.refresh()

    def _get_plotext_theme_name(self, app_theme_name: str) -> str:
        """Get the name of the Plotext theme to use.

        If the theme is set to `"auto"` the theme name will be of the form
        `textual-auto-<app_theme_name>`. Otherwise the theme name will be the
        Plotext theme name, which will be directly set on the plot..

        Args:
            app_theme_name: The name of the theme from the Textual app.
        """
        if self.theme == "auto":
            return f"textual-auto-{app_theme_name}"
        else:
            return self.theme



from textual.app import ComposeResult
from textual.binding import Binding
from textual.widgets import Static

from bagels.components.datatable import DataTable
from bagels.components.indicators import EmptyIndicator
from bagels.config import CONFIG
from bagels.forms.person_forms import PersonForm
from bagels.managers.persons import (
    delete_person,
    get_person_by_id,
    get_persons_with_net_due,
    update_person,
)
from bagels.modals.confirmation import ConfirmationModal
from bagels.modals.input import InputModal


class People(Static):
    COLUMNS = ("Name", "Net due")

    BINDINGS = [
        Binding(CONFIG.hotkeys.edit, "edit_person", "Edit"),
        Binding(CONFIG.hotkeys.delete, "delete_person", "Delete"),
    ]

    can_focus = True

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="people-container", classes="module-container"
        )
        super().__setattr__("border_title", "People")
        self.current_row = None

    # --------------- Hooks -------------- #

    def on_mount(self) -> None:
        self.rebuild()

    def on_data_table_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        if event.row_key:
            self.current_row = event.row_key.value

    # region Builders
    # ------------- Builders ------------- #

    def rebuild(self) -> None:
        table: DataTable = self.query_one("#people-table")
        empty_indicator: Static = self.query_one(".empty-indicator")

        table.clear()
        if not table.columns:
            table.add_columns(*self.COLUMNS)

        people = get_persons_with_net_due()
        if people:
            for person in people:
                table.add_row(person.name, person.due, key=person.id)
            table.zebra_stripes = True

        empty_indicator.display = not people

    # region Helpers
    # -------------- Helpers ------------- #

    def _notify_no_select(self) -> None:
        self.app.notify(
            title="Error",
            message="A person must be selected for this action.",
            severity="error",
            timeout=2,
        )

    # region callbacks
    # ------------- Callbacks ------------ #

    def action_delete_person(self) -> None:
        if not self.current_row:
            self._notify_no_select()
            return

        def check_delete(result) -> None:
            if result:
                try:
                    delete_person(self.current_row)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                self.rebuild()

        person = get_person_by_id(self.current_row)

        self.app.push_screen(
            ConfirmationModal(
                f"Are you sure you want to delete person '{person.name}'? Existing records with this person will not be affected."
            ),
            check_delete,
        )

    def action_edit_person(self) -> None:
        if not self.current_row:
            self._notify_no_select()
            return

        def check_result(result) -> None:
            if result:
                try:
                    update_person(self.current_row, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Person {result['name']} updated",
                        severity="information",
                        timeout=3,
                    )
                    self.rebuild()

        filled_form = PersonForm().get_filled_form(self.current_row)
        self.app.push_screen(
            InputModal("Edit Person", filled_form), callback=check_result
        )

    # region View
    # --------------- View --------------- #
    def compose(self) -> ComposeResult:
        yield DataTable(
            id="people-table",
            cursor_type="row",
            cursor_foreground_priority=True,
        )
        yield EmptyIndicator("No people")



import copy
from datetime import datetime

from rich.text import Text

from bagels.forms.form import Form, FormField, Option, Options
from bagels.managers.record_templates import get_transfer_templates
from bagels.models.record import Record


class TransferForm:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    # ------------ Blueprints ------------ #

    FORM = Form(
        fields=[
            FormField(
                placeholder="Label",
                title="Label / Template name",
                key="label",
                type="autocomplete",
                options=Options(),
                autocomplete_selector=False,
                is_required=True,
            ),
            FormField(
                title="Amount",
                key="amount",
                type="number",
                placeholder="0.00",
                min=0,
                is_required=True,
            ),
            FormField(
                title="Date",
                key="date",
                type="dateAutoDay",
                placeholder="dd (mm) (yy)",
            ),
        ]
    )

    TEMPLATE_FORM = Form(
        fields=[
            FormField(
                title="Label",
                key="label",
                type="string",
                placeholder="Label",
                is_required=True,
            ),
            FormField(
                title="Amount",
                key="amount",
                type="number",
                placeholder="0.00",
                min=0,
                is_required=True,
            ),
        ]
    )

    # ----------------- - ---------------- #

    def __init__(self, isTemplate: bool = False, defaultDate: str = None):
        self.isTemplate = isTemplate
        self.defaultDate = defaultDate
        self._populate_form_options()

    # region Helpers
    # -------------- Helpers ------------- #

    def _populate_form_options(self):
        templates = get_transfer_templates()
        self.FORM.fields[0].options = Options(
            items=[
                Option(
                    text=template.label,
                    value=template.id,
                    postfix=Text(f"{template.amount}", style="yellow"),
                )
                for template in templates
            ]
        )
        self.FORM.fields[2].default_value = self.defaultDate

    # region Builders
    # ------------- Builders ------------- #

    def get_filled_form(self, record: Record) -> Form:
        """Return a copy of the form with values from the record"""
        filled_form = copy.deepcopy(
            self.FORM if not self.isTemplate else self.TEMPLATE_FORM
        )

        if not record.isTransfer:
            return filled_form, []

        for field in filled_form.fields:
            fieldKey = field.key
            value = getattr(record, fieldKey)

            match fieldKey:
                case "date":
                    # if value is this month, simply set %d, else set %d %m %y
                    if value.month == datetime.now().month:
                        field.default_value = value.strftime("%d")
                    else:
                        field.default_value = value.strftime("%d %m %y")
                case "label":
                    field.default_value = str(value) if value is not None else ""
                    field.type = "string"  # disable autocomplete
                case _:
                    field.default_value = str(value) if value is not None else ""

        return filled_form

    def get_form(self, hidden_fields: dict = {}):
        """Return the base form with default values"""
        form = copy.deepcopy(self.FORM if not self.isTemplate else self.TEMPLATE_FORM)
        for field in form.fields:
            key = field.key
            if key in hidden_fields:
                field.type = "hidden"
                if isinstance(hidden_fields[key], dict):
                    field.default_value = hidden_fields[key]["default_value"]
                    field.default_value_text = hidden_fields[key]["default_value_text"]
                else:
                    field.default_value = hidden_fields[key]
        return form



from datetime import datetime

from bagels.forms.form import Form
from bagels.forms.record_forms import RecordForm
from bagels.managers.persons import (
    get_person_by_id,
    update_person,
)
from bagels.managers.record_templates import create_template_from_record
from bagels.managers.records import (
    create_record,
    create_record_and_splits,
    delete_record,
    get_record_by_id,
    update_record,
    update_record_and_splits,
)
from bagels.managers.splits import get_split_by_id, update_split
from bagels.modals.confirmation import ConfirmationModal
from bagels.modals.input import InputModal
from bagels.modals.record import RecordModal
from bagels.modals.transfer import TransferModal
from bagels.utils.format import format_date_to_readable


class RecordCUD:
    def action_new(self) -> None:
        def check_result(result) -> None:
            if result:
                try:
                    create_record_and_splits(result["record"], result["splits"])
                    if result["createTemplate"]:
                        create_template_from_record(result["record"])
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Record created {'and template created' if result['createTemplate'] else ''}",
                        severity="information",
                        timeout=3,
                    )
                    self.page_parent.rebuild(templates=result["createTemplate"])

        date = format_date_to_readable(self.page_parent.mode["date"])
        account_name = self.page_parent.mode["accountId"]["default_value_text"]
        type = "Income" if self.page_parent.mode["isIncome"] else "Expense"
        self.app.push_screen(
            RecordModal(
                f"New {type} on {account_name} for {date}",
                form=RecordForm().get_form(self.page_parent.mode),
                splitForm=Form(),
                date=self.page_parent.mode["date"],
            ),
            callback=check_result,
        )

    def action_edit(self) -> None:
        if not (hasattr(self, "current_row") and self.current_row):
            self.app.notify(
                title="Error", message="Nothing selected", severity="error", timeout=2
            )
            self.app.bell()
            return
        # ----------------- - ---------------- #
        type = self.current_row.split("-")[0]
        id = self.current_row.split("-")[1]

        # ----------------- - ---------------- #
        def check_result_records(result) -> None:
            if result:
                try:
                    if result.get("record"):  # if not editing a transfer:
                        update_record_and_splits(id, result["record"], result["splits"])
                    else:
                        update_record(id, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message="Record updated",
                        severity="information",
                        timeout=3,
                    )
                    self.page_parent.rebuild()
            else:
                self.app.notify(
                    title="Discarded",
                    message="Record not updated",
                    severity="warning",
                    timeout=3,
                )

        def check_result_person(result) -> None:
            if result:
                try:
                    update_person(id, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message="Person updated",
                        severity="information",
                        timeout=3,
                    )
                    self.page_parent.rebuild()
            else:
                self.app.notify(
                    title="Discarded",
                    message="Person not updated",
                    severity="warning",
                    timeout=3,
                )

        # ----------------- - ---------------- #
        match type:
            case "r":
                record = get_record_by_id(id)
                if not record:
                    self.app.notify(
                        title="Error",
                        message="Record not found",
                        severity="error",
                        timeout=2,
                    )
                    return
                if record.isTransfer:
                    self.app.push_screen(
                        TransferModal(title="Edit transfer", record=record),
                        callback=check_result_records,
                    )
                else:
                    filled_form, filled_splits = RecordForm().get_filled_form(record.id)
                    self.app.push_screen(
                        RecordModal(
                            "Edit Record",
                            form=filled_form,
                            splitForm=filled_splits,
                            isEditing=True,
                        ),
                        callback=check_result_records,
                    )
            case "s":
                split = get_split_by_id(id)
                if split.isPaid:
                    split_data = {"accountId": None, "isPaid": False, "paidDate": None}
                    update_split(id, split_data)
                    self.app.notify(
                        title="Reverted split",
                        message="Marked this split as unpaid",
                        severity="information",
                        timeout=3,
                    )
                else:
                    split_data = {
                        "accountId": self.page_parent.mode["accountId"][
                            "default_value"
                        ],
                        "isPaid": True,
                        "paidDate": datetime.now(),
                    }
                    update_split(id, split_data)
                    self.app.notify(
                        title="Completed split",
                        message=f"With account {self.page_parent.mode['accountId']['default_value_text']} today",
                        severity="information",
                        timeout=3,
                    )
                self.page_parent.rebuild()
            case "p":
                person = get_person_by_id(id)
                if not person:
                    self.app.notify(
                        title="Error",
                        message="Person not found",
                        severity="error",
                        timeout=2,
                    )
                    return
                self.app.push_screen(
                    InputModal(
                        "Edit Person", form=self.person_form.get_filled_form(person.id)
                    ),
                    callback=check_result_person,
                )
            case _:
                pass

    def action_delete(self) -> None:
        if not (hasattr(self, "current_row") and self.current_row):
            self.app.notify(
                title="Error", message="Nothing selected", severity="error", timeout=2
            )
            self.app.bell()
            return
        # ----------------- - ---------------- #
        type = self.current_row.split("-")[0]
        id = self.current_row.split("-")[1]

        if type == "s":
            self.app.notify(
                title="Error",
                message="You cannot delete or add splits to a record after creation.",
                severity="error",
                timeout=2,
            )
            return

        # ----------------- - ---------------- #
        def check_delete(result) -> None:
            if result:
                delete_record(id)
                self.app.notify(
                    title="Success",
                    message="Record deleted",
                    severity="information",
                    timeout=3,
                )
                self.page_parent.rebuild()

        # ----------------- - ---------------- #
        match type:
            case "r":
                self.app.push_screen(
                    ConfirmationModal("Are you sure you want to delete this record?"),
                    callback=check_delete,
                )
            case "s":
                self.app.push_screen(
                    ConfirmationModal("Are you sure you want to delete this split?"),
                    callback=check_delete,
                )
            case _:
                pass

    def action_new_transfer(self) -> None:
        def check_result(result) -> None:
            if result:
                try:
                    create_record(result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message="Record created",
                        severity="information",
                        timeout=3,
                    )
                    self.page_parent.rebuild()

        self.app.push_screen(
            TransferModal(
                title="New transfer",
                defaultDate=self.page_parent.mode["date"].strftime("%d"),
            ),
            callback=check_result,
        )
